// Filename: cmd_trace.c / cmd_trace.so
// Author: David Ljung Madison <DaveSource.com>
// See License:  http://MarginalHacks.com/License/
// Description:  LD_PRELOAD exec/command tracer
//   Can show you all the exec calls (with stdout/stderr redirects)
//   for scripts as well as any child scripts and their children etc..
//
//   Also shows ENV changes (though not deletions)
//   Note that multiple levels may each add new ENV variables, so also
//   add in 'env var=val' from prev commands.
//   A future script may separate this and include unsetenv as well...
//
//   Great for debugging script wrappers with many levels of wrappers.
//
// Compilation:
// % cc -Wall -fPIC -DPIC -ldl -ggdb   -c -o cmd_trace.o cmd_trace.c
// % cc -shared -o cmd_trace.so cmd_trace.o -ldl
// Run:
// % env LD_PRELOAD=cmd_trace.so <my command> <my args....>
// Output:
//   Look for '[EXEC_TRACE]' lines in stderr or send to a log
//
// Implements:
//   execv, execvp, execve, execvpe
//   fexecve, execveat
//   execl, execlp, execle, execlpe
//   posix_spawn, posix_spawnp
//
// Inspired by Billy Holmes @gonoph.net

#define _GNU_SOURCE
#include <dlfcn.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <spawn.h>

// For file exists checking
#include <sys/stat.h>

// For getting stdout/stderr paths
#include <unistd.h>
#include <limits.h>

// Avoid collisions or unintentional overwriting of any calls
#define NAMESPACE(x) (__cmd_trace_##x)

FILE * NAMESPACE(log) = NULL;
int NAMESPACE(env_line) = 0;
int NAMESPACE(no_bin) = 0;
int NAMESPACE(show_pipes) = 0;

char *NAMESPACE(prefix) = "[EXEC_TRACE] ";
#define OUT_PREFIX if (NAMESPACE(prefix)!=NULL) fputs(NAMESPACE(prefix), NAMESPACE(log));

// -------------------------
// Quoting
// -------------------------
// Output safe shell quote of a string
void NAMESPACE(quote)(const char *str) {
	const char *p = str;

	// Do we need to quote?
	int safe=1;
	while (*p != '\0') {
		const char ch = *p++;
		// Safe set of chars:  a-zA-Z0-9,._+:@%/=-
		if (ch >= 'a' && ch<= 'z') continue;
		if (ch >= 'A' && ch<= 'Z') continue;
		if (ch >= '0' && ch<= '9') continue;
		if (ch == ',' || ch == '.' || ch == '_' || ch == '+' || ch == ':'
			|| ch == '@' || ch == '%' || ch == '/' || ch == '-' || ch == '=') continue;
		safe=0;
		break;
	}

	// No quoting needed, just output
	if (safe==1) {
		fputs(str, NAMESPACE(log));
		return;
	}

	// We need to quote.
	// Wrap with ['] and then wrap ['] with ["'"] (with end/begin ' around)
	p = str;
	fputs("'", NAMESPACE(log));
	while (*p != '\0') {
		if (*p == '\'') {
			fputs("'\"'\"'", NAMESPACE(log));
		} else {
			fputc(*p, NAMESPACE(log));
		}
		p++;
	}
	fputs("'", NAMESPACE(log));
}

// -------------------------
// Enviroment variable changes
// -------------------------
void NAMESPACE(free_charv)(char **charv) {
	if (charv==NULL) return;
	for (char **it = charv; it && *it; ++it) { free(*it); }
	free(charv);
	charv = NULL;
}

char **NAMESPACE(orig_env) = NULL;
void NAMESPACE(save_env)() {
	int num = 0;

	// If environment was saved, free it
	NAMESPACE(free_charv)(NAMESPACE(orig_env));

	if (environ == NULL) return;

	while (environ[num] != NULL) { ++num; }
	NAMESPACE(orig_env) = malloc((num+1)*sizeof(char *));
	if (NAMESPACE(orig_env) != NULL) {
		num = 0;
		while (environ[num] != NULL) {
			NAMESPACE(orig_env)[num] = strdup(environ[num]);
			++num;
		}
		NAMESPACE(orig_env)[num] = NULL;
	}
}

int orig_env_has(const char *str) {
	int i = 0;
	while (NAMESPACE(orig_env)[i] != NULL) {
		if (strcmp(str, NAMESPACE(orig_env)[i])==0)
			return 1;
		++i;
	}
	return 0;
}

void env_changes(char * const *env) {
	if (env==NULL) return;
	if (NAMESPACE(orig_env)==NULL) return;
	int num = 0;
	int found = 0;
	while (env[num] != NULL) {
		if (orig_env_has(env[num])==0) {
			if (found++==0) {
				OUT_PREFIX;
				fputs(NAMESPACE(env_line)==1 ? "% env" : "%", NAMESPACE(log));
			}

			// Check if they've changed LD_PRELOAD!
			char *eq = index(env[num],'=');
			if (eq!=NULL && strncmp("LD_PRELOAD", env[num], eq-env[num])==0) {
				fprintf(NAMESPACE(log),"\n[EXEC_TRACE] WARN: Script is changing LD_PRELOAD - breaking cmd_trace functionality\n");
			}

			// Output, possibly quoting the value
			int chnum = 0;
			fputc(' ', NAMESPACE(log));
			while (env[num][chnum] != 0 && env[num][chnum] != '=') {
				fputc(env[num][chnum++], NAMESPACE(log));
			}
			fputc('=', NAMESPACE(log));
			if (env[num][chnum] != 0 && env[num][chnum+1] != 0)
				NAMESPACE(quote)(env[num]+chnum+1);
		}
		++num;
	}
	if (found>0 && NAMESPACE(env_line)!=1) fputs("\n",NAMESPACE(log));

	// Update our copy of the environment
	// We could have done this piecemeal in the loop above, but this is simple and fast
	NAMESPACE(save_env)();
}

// -------------------------
// Output Commands
// -------------------------
void NAMESPACE(setup_log)() {
	// Setup log or use stderr
	const char *logpath = getenv("EXEC_TRACE_LOG");
	NAMESPACE(log) = stderr;
	if (logpath != NULL) {
		NAMESPACE(log) = fopen(logpath,"a");	// Append, because we could be a child
		if (NAMESPACE(log) == NULL) {
			fprintf(stderr,"[EXEC_TRACE] ERROR: Couldn't write log %s\n",logpath);
			NAMESPACE(log) = stderr;
		} else {
			// Turn off buffering for the logfile
			setvbuf(NAMESPACE(log), (char *)NULL, _IONBF, 0 );
			NAMESPACE(prefix) = NULL;
		}
	}
}

// Quote and output the env/args/stdout/err for execv calls
int NAMESPACE(is_bin)(const char *path) {
	struct stat tmpbuf;

	if (path == NULL) return 0;

	// Path:  /bin/ls -> is_bin
	if (strncmp("/bin/",path,5) == 0) return 1;
	// Path:  ls -> check /bin/ls -> is_bin
	char *binpath = (char*) malloc((strlen(path)+7)*(sizeof(char)));
	if (binpath != NULL) {
		strcpy(binpath, "/bin/"); strcat(binpath, path);
		if (stat(binpath,&tmpbuf) == 0) return 1;
	}

	return 0;
}

ssize_t NAMESPACE(read_fd)(int fd, char buf[], ssize_t sz) {
	char path[1024];
	strcpy(path,"/proc/self/fd/");
	sprintf(path+14, "%d", fd);
	buf[0] = 0;	// In case of failure
	return readlink(path,buf,sz-1);
}

void NAMESPACE(out)(const char *path, char *const argv[], char *const envp[]) {
	if (NAMESPACE(no_bin) == 1 && NAMESPACE(is_bin)(path) == 1) return;

	NAMESPACE(setup_log)();

	const char *ri = rindex(path,'/');
	const char *pathfile = (ri == NULL) ? path : ri + 1;

	if (!argv[0] || (strcmp(path, argv[0]) != 0 && strcmp(pathfile, argv[0]) != 0)) {
		OUT_PREFIX;
		fputs("WARN: $0 is not path: ", NAMESPACE(log));
		NAMESPACE(quote)(path);
		fputs("\n", NAMESPACE(log));
	}

	// Output changed environment vars
	env_changes((envp==NULL) ? environ : envp);
	// Output changed environment vars

	if (NAMESPACE(env_line)==0) {
		OUT_PREFIX;
		fputs("%", NAMESPACE(log));
	}

	// Output the command, quoting as necessary
	int i = 0;
	while (argv[i] != NULL) {
		fputc(' ', NAMESPACE(log));
		NAMESPACE(quote)(argv[i++]);
	}

	fputs("\n", NAMESPACE(log));

	// Get stdout/stderr redirects
	char buf[PATH_MAX];
	ssize_t len = NAMESPACE(read_fd)(1,buf,sizeof(buf));
	if (len != -1 && (len<5 || strncmp(buf,"/dev/",5)!=0)) {
		if (NAMESPACE(show_pipes)==1 || len<5 || strncmp(buf,"pipe:",5)!=0) {
			buf[len]='\0';
			OUT_PREFIX;
			fprintf(NAMESPACE(log),"stdout: %s\n",buf);
		}
	}
	len = NAMESPACE(read_fd)(2,buf,sizeof(buf));
	if (len != -1 && (len<5 || strncmp(buf,"/dev/",5)!=0)) {
		if (NAMESPACE(show_pipes)==1 || len<5 || strncmp(buf,"pipe:",5)!=0) {
			buf[len]='\0';
			OUT_PREFIX;
			fprintf(NAMESPACE(log),"stderr: %s\n",buf);
		}
	}
}

// -------------------------
// Wrappers
// -------------------------
int (*NAMESPACE(orig_execv))(const char *__path, char *const __argv[]);
int execv (const char *__path, char *const __argv[]) {
	NAMESPACE(out)(__path, __argv, NULL);
	return NAMESPACE(orig_execv)(__path, __argv);
}

int (*NAMESPACE(orig_execvp))(const char *__path, char *const __argv[]);
int execvp (const char *__path, char *const __argv[]) {
	NAMESPACE(out)(__path, __argv, NULL);
	return NAMESPACE(orig_execvp)(__path, __argv);
}

int (*NAMESPACE(orig_execve))(const char *__path, char *const __argv[], char *const __envp[]);
int execve (const char *__path, char *const __argv[], char *const __envp[]) {
	NAMESPACE(out)(__path, __argv, __envp);
	return NAMESPACE(orig_execve)(__path, __argv, __envp);
}

int (*NAMESPACE(orig_execvpe))(const char *__path, char *const __argv[], char *const __envp[]);
int execvpe(const char *__path, char *const __argv[], char *const __envp[]) {
	NAMESPACE(out)(__path, __argv, __envp);
	return NAMESPACE(orig_execvpe)(__path, __argv, __envp);
}

int (*NAMESPACE(orig_fexecve))(int fd, char *const __argv[], char *const __envp[]);
int fexecve (int fd, char *const __argv[], char *const __envp[]) {
	char buf[PATH_MAX];
	NAMESPACE(read_fd)(fd,buf,sizeof(buf));	// Get the path from the fd
	NAMESPACE(out)(buf, __argv, __envp);
	return NAMESPACE(orig_fexecve)(fd, __argv, __envp);
}

int (*NAMESPACE(orig_execveat))(int dirfd, char *const __path, char *const __argv[], char *const __envp[], int flags);
int execveat (int dirfd, char *const __path, char *const __argv[], char *const __envp[], int flags)
{
	NAMESPACE(out)(__path, __argv, __envp);
	return NAMESPACE(orig_execveat)(dirfd, __path, __argv, __envp, flags);
}

char **NAMESPACE(buildargv)(const char *arg, va_list *args)
{
	char **argv = NULL;
	size_t size = 0;
	for (const char *it = arg; it; it = va_arg(*args, const char *)) {
		argv = realloc(argv, (size + 2) * sizeof(char *));
		if (!argv) {
			fprintf(stderr,"[exec_trace] FATAL ERROR: Could not malloc buildargv\n");
			exit(-1);
		}
		argv[size] = strdup(it);
		if (!argv[size]) {
			fprintf(stderr,"[exec_trace] FATAL ERROR: Could not malloc buildargv string\n");
			exit(-1);
		}
		argv[++size] = NULL;
	}
	return argv;
}

int execl(const char *__path, const char *__arg, ...)
{
	va_list args;
	va_start(args,__arg);
	char **argv = NAMESPACE(buildargv)(__arg, &args);
	va_end(args);
	NAMESPACE(out)(__path, argv, NULL);
	return NAMESPACE(orig_execv)(__path, argv);
}

int execlp(const char *__path, const char *__arg, ...)
{
	va_list args;
	va_start(args,__arg);
	char **argv = NAMESPACE(buildargv)(__arg, &args);
	va_end(args);
	NAMESPACE(out)(__path, argv, NULL);
	return NAMESPACE(orig_execvp)(__path, argv);
}

int execle(const char *__path, const char *__arg, ...)
{
	va_list args;
	va_start(args,__arg);
	char **argv = NAMESPACE(buildargv)(__arg, &args);
	char *const *envp = va_arg(args, char *const *);
	va_end(args);

	NAMESPACE(out)(__path, argv, envp);
	return NAMESPACE(orig_execve)(__path, argv, envp);
}

// Probably doesn't exist
int execlpe(const char *__path, const char *__arg, ...)
{
	va_list args;
	va_start(args,__arg);
	char **argv = NAMESPACE(buildargv)(__arg, &args);
	char *const *envp = va_arg(args, char *const *);
	va_end(args);

	NAMESPACE(out)(__path, argv, envp);
	return NAMESPACE(orig_execvpe)(__path, argv, envp);
}

int (*NAMESPACE(orig_posix_spawn))(pid_t *pid, const char *path, const posix_spawn_file_actions_t *file_actions, const posix_spawnattr_t *attrp, char *const argv[], char *const envp[]);
int posix_spawn(pid_t *pid, const char *path, const posix_spawn_file_actions_t *file_actions, const posix_spawnattr_t *attrp, char *const argv[], char *const envp[])
{
	NAMESPACE(out)(path, argv, envp);
	return NAMESPACE(orig_posix_spawn)(pid, path, file_actions, attrp, argv, envp);
}

// -------------------------
// Setup
// -------------------------
// On load we save the original pointers for the exec calls
void __attribute__ ((constructor)) my_init(void) {
	//fputs("[Loading cmd_trace]\n", NAMESPACE(log));

	// Copy the original exec function pointers
	NAMESPACE(orig_execv) = dlsym(RTLD_NEXT, "execv");
	NAMESPACE(orig_execvp) = dlsym(RTLD_NEXT, "execvp");
	NAMESPACE(orig_execve) = dlsym(RTLD_NEXT, "execve");
	NAMESPACE(orig_execvpe) = dlsym(RTLD_NEXT, "execvpe");
	NAMESPACE(orig_fexecve) = dlsym(RTLD_NEXT, "fexecve");
	NAMESPACE(orig_execveat) = dlsym(RTLD_NEXT, "execveat");
	NAMESPACE(orig_posix_spawn) = dlsym(RTLD_NEXT, "posix_spawn");

	// Copy the environ
	NAMESPACE(save_env)();

	// Options:  -env and -bin
	const char *envvar = getenv("EXEC_TRACE_ENV");
	if (envvar != NULL && strcmp("0",envvar) != 0) NAMESPACE(env_line) = 1;

	const char *binvar = getenv("EXEC_TRACE_BIN");
	if (binvar != NULL && strcmp("0",binvar) != 0) NAMESPACE(no_bin) = 1;

	const char *pipevar = getenv("EXEC_TRACE_PIPE");
	if (pipevar != NULL && strcmp("0",pipevar) != 0) NAMESPACE(show_pipes) = 1;
}

